% source code for example in Appendix B.1 of the paper
% written and tested in Octave, Version 5.2.0
clear all 

%rand ("seed", 12345)   % uncomment this line to get repeatable results (specifically reproduce the numbers in the paper)

% Construct a game with randomly chosen equilibria
n = 5;  % number of strategies for player 1
m = 6;  % number of strategies for player 2
% number of desired equilibria
k1 = 2; 
k2 = 3;
% pick some random vectors to be equilibrium strategies
% for player 1
p = zeros(k1, n);
for i = 1:k1
  aux = rand(1, n);
  p(i,:) = aux / sum(aux);
endfor
% likewise, pick some random vectors to be 
% equilibrium strategies for player 2
q = zeros(k2, m);
for i = 1:k2
  aux = rand(1, m);
  q(i,:) = aux / sum(aux);
endfor
X = null(p)' 
Y = null(q)'

% now, pick a random matrix Z to "connect" X and Y
Z = rand(size(X,1), size(Y,1))

% this is the payoff for the first goal
A1 = X'*Z*Y

% we have (verifiably, up to numeric precision)
% X*p' == 0
% Y*q' == 0
% furthermore:
% q(i,:) * A1 == 0     for all i
% A1 * (p(i,:))' == 0  for all i

% this is X_(k-1)
eq11 = p;
          
% this is Y_(k-1)
eq12 = q;

nk = size(eq11,1); % = size(eq12,1)
          
% do the pairwise combination of equilibria to construct the next stage game
A2 = randi(5, size(A1)) % we pick a randomly payoff structure A2 for the second goal

% compute the payoff for the second goal (equation (1) in the paper)
B2 = eq11 * A2 * eq12';

% now, let us pick an arbitrary but fixed equilibrium in B2
% ...and compute the equilibrium using the standard linear programming
% LP for player 1
A = [ -B2'                 ones(size(B2,2),1) ;
      ones(1,size(B2,1))  0                   ];
c = [zeros(size(B2,1),1); 1];
b = [zeros(size(B2,2),1); 1];
[xmax, fmax, status2, extra2] = glpk(c, A, b, [], [], 
    [repmat("U", [1 size(b,1)-1]) "S"], 
    [repmat("C", [1 size(c,1)])], -1);
% LP for player 2
[ymin, fmin, status, extra] = glpk(b, A', c, [], [], 
    [repmat("L", [1 size(c,1)-1]) "S"], 
    [repmat("C", [1 size(b,1)])], 1);

alpha = xmax(1:(size(c)-1))
beta = ymin(1:(size(b)-1))

% x1mix = x_k', y1mix = y_k' (in the paper)
x1mix = alpha' * eq11; %  = (alpha_k*)^T * X_(k-1)
y1mix = beta' * eq12; %  = Y_(k-1)^T * beta_k*

%%% Finally, for a humble/naive trial verification that (x1mix, y1mix) is really
%%% lex-optimal; note that a systematic search for a better solution would be just 
%%% what was done above anyway

% first, let us take player 1's view:
trials = 10000
disp('trial verification for player 1')
v1 = x1mix * A1 * y1mix';    % will always equal the (same) saddle point value 
v2 = x1mix * A2 * y1mix'; 

n = size(A1,1);
for i=1:trials 
  if (i <= n) 
    % try pure strategies as alternative first; are they optimal?
    xalt = zeros(1, n);
    xalt(i) = 1;
  else 
    % now, try random strategies; do we find a better one that computed above?
    xalt = rand(1, n);
    xalt = xalt / sum(xalt);
  endif
  v1alt = xalt * A1 * y1mix';
  v2alt = xalt * A2 * y1mix';
  
  if (lexgt([v1alt v2alt], [v1 v2] + 1e-6))  % tolerance 1e-6 to account for numerical roundoff errors 
    disp('found a better solution here:')  % if the computation is correct, this message should never show up
    disp([v1 v1alt])
    break
  endif
endfor 

%%%%% now, from player 2's perspective
disp('trial verifications for player 2')
n = size(A1, 2);
for i=1:trials 
  if (i <= n) 
    % as before, let us try pure strategies first
    yalt = zeros(1, n);
    yalt(i) = 1;
  else 
    % and also try randomly mixed strategies to look for a better result
    yalt = rand(1, n);
    yalt = yalt / sum(yalt);
  endif
  v1alt = x1mix * A1 * yalt';
  v2alt = x1mix * A2 * yalt';
  % lex-<=-predicate computed by "a < c" <=> "-a > -c"
  if (lexgt(-[v1alt v2alt], -[v1 v2] + 1e-6))   % include round-off errors
    disp('found a better solution here:') % again, this should never be displayed if everything work's the right way
    disp([v1 v1alt])
    break
  endif
endfor 

disp('done')
